package tud.st.dpf.robot.simulator;

import java.util.ArrayList;
import java.util.List;

import tud.st.dpf.robot.simulator.participants.ChargingStation;
import tud.st.dpf.robot.simulator.participants.Machine;
import tud.st.dpf.robot.simulator.participants.Order;
import tud.st.dpf.robot.simulator.participants.Person;
import tud.st.dpf.robot.simulator.participants.Position;
import tud.st.dpf.robot.simulator.participants.Product;
import tud.st.dpf.robot.simulator.participants.Robot;
import tud.st.dpf.robot.simulator.participants.Storage;

public class Simulator {
	
	private static Simulator instance;

	private Thread simulatorThread;
	private List<Robot> robots;
	private List<Machine> machines;
	private List<Person> persons;
	private ChargingStation chargingStation;
	private Storage storage;

	private List<Order> orders;
	private Order currentOrder;
	private List<Product> productsLeftInCurrentOrder;
	private long orderStartTime = 0;
	private int itemsToProduce;
	private int itemsProduced = 0;

	public static Simulator getInstance(){
		if(instance == null)
			instance = new Simulator();
		
		return instance;
	}
	
	public void initialize(List<Robot> robots, List<Machine> machines, List<Person> persons){
		this.robots = robots;
		this.persons = persons;
		this.machines = machines;
		this.orders = new ArrayList<Order>();
		chargingStation = new ChargingStation("Charging Station");
		storage = new Storage("Storage");
		this.productsLeftInCurrentOrder = new ArrayList<Product>();
		
		for(Robot r : robots)
			r.setCurrentPosition(chargingStation);
	}
	
	private Simulator() {
	}

	public void start() {
		simulatorThread = new Thread(new Runnable() {

			@Override
			public void run() {
				while (true) {
					for(Person p : persons)
						p.work();
					
					if (currentOrder == null)
						pickOrder();

					//Create new Tasks for creating new products
					if (currentOrder != null) {
						Product p = pickNextProduct();

						if (p != null) {
							Robot r = pickFreeRobot(p);
							if(r != null){
								Position nextPos = getPositionForProduct(p);
								r.driveTo(nextPos, r.getMaxSpeed());
							}
							else{
								productsLeftInCurrentOrder.add(p);
							}
						}
					}
					
					//Those robots who arrived at machines have to load the products
					List<Robot> robotsArrivedAtMachines = getRobotsArrivedAtMachines();
					for(Robot r : robotsArrivedAtMachines){
						takeProductAndDriveToStorage(r);
					}

					//Those robots who arrived at the storage should put the product in the storage and drive back to the charging station
					List<Robot> robotsArrivedAtTheStorage = getRobotsArrivedAtStorage();
					for(Robot r : robotsArrivedAtTheStorage)
						dropProductAndDriveToChargingStation(r);
					
					if (currentOrder != null && productsLeftInCurrentOrder.size() == 0 && itemsProduced == itemsToProduce && allRobotsInChargingStation())
						finishCurrentOrder();
				}
			}

			private boolean allRobotsInChargingStation() {
				for(Robot r : robots)
					if(!r.isBroken() && !(r.getCurrentPosition() instanceof ChargingStation))
						return false;
				
				return true;
			}
		});

		simulatorThread.start();
	}

	public void stop() {
		simulatorThread = null;
	}

	public void processOrder(Order o) {
		orders.add(o);
	}

	protected List<Robot> getRobotsArrivedAtStorage() {
		List<Robot> result = new ArrayList<Robot>();
		
		for(Robot r : robots)
			if(r.getCurrentPosition() == storage)
				result.add(r);
		
		return result;
	}

	protected void takeProductAndDriveToStorage(Robot r) {
		Machine m = (Machine)r.getCurrentPosition();
		Product newProduct = m.produce();
		itemsProduced ++;
		r.take(newProduct);
		r.driveTo(storage, r.getMaxSpeed());
	}

	protected List<Robot> getRobotsArrivedAtMachines() {
		List<Robot> result = new ArrayList<Robot>();
		for(Robot r : robots){
			if(r.getCurrentPosition() != null && r.getCurrentPosition() instanceof Machine){
				result.add(r);
			}
		}
		
		return result;
	}

	protected void dropProductAndDriveToChargingStation(Robot r) {
		Product p = r.unloadObject();
		storage.store(p);
		
		r.driveTo(chargingStation, r.getMaxSpeed());
	}

	protected Product pickNextProduct() {
		if(productsLeftInCurrentOrder.size() > 0){
			Product result = productsLeftInCurrentOrder.get(0);
			productsLeftInCurrentOrder.remove(result);
			return result;
		}
		
		return null;
	}

	protected void pickOrder() {
		if (orders.size() > 0) {
			currentOrder = orders.get(0);
			orders.remove(currentOrder);
			productsLeftInCurrentOrder = currentOrder.getProducts();
			orderStartTime = System.currentTimeMillis();
			itemsToProduce = currentOrder.getProducts().size();
		}
	}

	protected Robot pickFreeRobot(Product p) {
		for (Robot r : robots)
			if (r.getCurrentPosition() != null && r.getCurrentPosition() == chargingStation && r.canTakeTask(p))
				return r;

		return null;
	}
	
	protected Position getPositionForProduct(Product p){
		for(Machine m : machines)
			if(m.produces(p))
				return m;
		
		return null;
	}

	protected void finishCurrentOrder() {
		currentOrder = null;
		productsLeftInCurrentOrder.clear();
		
		double seconds = (System.currentTimeMillis() - orderStartTime) / 1000.0;
		int brokenRobots = 0;
		
		for(Robot r : robots)
			if(r.isBroken())
				brokenRobots++;
		
		int itemsLost = itemsToProduce - storage.getStoredItemCount();
		
		System.out.println("----------- ORDER FINISHED -----------");
		System.out.println("Time: " + seconds + "s");
		System.out.println("Broken Robots: " + brokenRobots);
		System.out.println("Items Lost: " + itemsLost);
	}

}
